@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 1999-2022. All Rights Reserved.
+%% Copyright Ericsson AB 1999-2024. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -45,6 +45,36 @@
 -export([init/1, handle_call/3, handle_cast/2, handle_info/2, 
 	 terminate/2, code_change/3]).
 
+-opaque connection_reference() :: pid().
+-type col_name()             :: string().
+-type row()                  :: tuple().
+-type value()                :: null | term().
+-type selected()             :: {selected, [col_name()], [row()]}.
+-type updated()              :: {updated, n_rows()}.
+-type n_rows()               :: integer().
+-type odbc_data_type()       ::  sql_integer | sql_smallint | sql_tinyint |
+                                 {sql_decimal, Precision::integer(), Scale::integer()} |
+                                 {sql_numeric, Precision::integer(), Scale::integer()} |
+                                 {sql_char, Size::integer()} |
+                                 {sql_wchar, Size::integer()} |
+                                 {sql_varchar, Size::integer()} |
+                                 {sql_wvarchar, Size::integer()}|
+                                 {sql_float, Precision::integer()} |
+                                 {sql_wlongvarchar, Size::integer()} |
+                                 {sql_float, Precision::integer()} |
+                                 sql_real | sql_double | sql_bit | atom().
+-type common_reason()        :: connection_closed | extended_error() | term().
+-type extended_error()       :: {string(), integer(), term()}.
+
+-export_type([connection_reference/0,
+              col_name/0, row/0,
+              selected/0,
+              updated/0,
+              n_rows/0,
+              odbc_data_type/0,
+              common_reason/0,
+              extended_error/0]).
+
 %%--------------------------------------------------------------------------
 %% Internal state
 -record(state, {erlang_port,                 % The port to the c-program
@@ -75,23 +105,26 @@
 %%%  API
 %%%=========================================================================
 
+%%--------------------------------------------------------------------
+-spec start() -> ok | {error, Reason} when
+      Reason ::term().
+%%--------------------------------------------------------------------
+start() ->
+    application:start(odbc).
 
 %%--------------------------------------------------------------------
-%% Function: start([, Type]) -> ok
-%%
-%%  Type =  permanent | transient | temporary
+-spec start(Type) -> ok | {error, Reason} when
+      Type :: permanent | transient | temporary,
+      Reason ::term().
 %%
 %% Description: Starts the inets application. Default type
 %% is temporary. see application(3)
 %%--------------------------------------------------------------------
-start() -> 
-    application:start(odbc).
-
 start(Type) -> 
     application:start(odbc, Type).
 
 %%--------------------------------------------------------------------
-%% Function: stop() -> ok
+-spec stop() -> ok.
 %%
 %% Description: Stops the odbc application.
 %%--------------------------------------------------------------------
@@ -99,8 +132,19 @@ stop() ->
     application:stop(odbc).
 
 %%-------------------------------------------------------------------------
-%% connect(ConnectionStr, Options) -> {ok, ConnectionReferense} |
-%%                                    {error, Reason}
+-spec connect(ConnectionStr, Options) -> {ok, ConnectionReferense} |
+          {error, Reason} when
+      ConnectionStr :: string(),
+      Options :: [{auto_commit, on | off} |
+                  {timeout, erlang:timeout()} |
+                  {binary_strings, on | off} |
+                  {tuple_row, on | off} |
+                  {scrollable_cursors, on | off} |
+                  {trace_driver, on | off} |
+                  {extended_errors, on | off}],
+      ConnectionReferense :: connection_reference(),
+      Reason :: port_program_executable_not_found | common_reason().
+
 %% Description: Spawns an erlang control process that will open a port
 %%              to a c-process that uses the ODBC API to open a connection
 %%              to the database. 
@@ -119,7 +163,9 @@ connect(ConnectionStr, Options) when is_list(ConnectionStr), is_list(Options) ->
     end.
 
 %%--------------------------------------------------------------------------
-%% disconnect(ConnectionReferense) -> ok | {error, Reason}
+-spec disconnect(ConnectionReferense) -> ok | {error, Reason} when
+      ConnectionReferense :: connection_reference(),
+      Reason :: process_not_owner_of_odbc_connection | extended_error().
 %%                                    
 %% Description: Disconnects from the database and terminates both the erlang
 %%              control process and the database handling c-process. 
@@ -144,14 +190,23 @@ disconnect(ConnectionReference) when is_pid(ConnectionReference)->
     end. 
 	    
 %%--------------------------------------------------------------------------
-%% commit(ConnectionReference, CommitMode, <TimeOut>) -> ok | {error,Reason}
-%%                                    
-%% Description: Commits or rollbacks a transaction. Needed on connections
-%%              where automatic commit is turned off.  
+-spec commit(ConnectionReference, CommitMode) -> ok | {error, Reason} when
+      ConnectionReference :: connection_reference(),
+      CommitMode ::  commit | rollback,
+      Reason :: not_an_explicit_commit_connection |
+                process_not_owner_of_odbc_connection | common_reason().
 %%--------------------------------------------------------------------------
 commit(ConnectionReference, CommitMode) ->
     commit(ConnectionReference, CommitMode, ?DEFAULT_TIMEOUT).
 
+%%--------------------------------------------------------------------------
+-spec commit(ConnectionReference, CommitMode, TimeOut) -> ok | {error, Reason} when
+      ConnectionReference :: connection_reference(),
+      CommitMode ::  commit | rollback,
+      TimeOut :: erlang:timeout(),
+      Reason :: not_an_explicit_commit_connection |
+                process_not_owner_of_odbc_connection | common_reason().
+%%--------------------------------------------------------------------------
 commit(ConnectionReference, commit, infinity) 
   when is_pid(ConnectionReference) ->
     ODBCCmd = [?COMMIT_TRANSACTION, ?COMMIT],
@@ -173,16 +228,25 @@ commit(ConnectionReference, rollback, TimeOut)
     call(ConnectionReference, {commit, ODBCCmd}, TimeOut).
 
 %%--------------------------------------------------------------------------
-%% sql_query(ConnectionReference, SQLQuery, <TimeOut>) -> {updated, NRows} |
-%%			       {selected, ColNames, Rows} | {error, Reason} 
+-spec sql_query(ConnectionReference, SQLQuery) -> Result | {error, Reason}  when
+      ConnectionReference :: connection_reference(),
+      SQLQuery :: string(),
+      Result  :: updated() | selected(),
+      Reason  :: process_not_owner_of_odbc_connection | common_reason().
 %%                                    
-%% Description: Executes a SQL query. If it is a SELECT query the
-%%              result set is returned, otherwise the number of affected 
-%%       	rows are returned.
 %%--------------------------------------------------------------------------
 sql_query(ConnectionReference, SQLQuery) ->
     sql_query(ConnectionReference, SQLQuery, ?DEFAULT_TIMEOUT).
 
+%%--------------------------------------------------------------------------
+-spec sql_query(ConnectionReference, SQLQuery, TimeOut) -> Result | {error, Reason}  when
+      ConnectionReference :: connection_reference(),
+      SQLQuery :: string(),
+      TimeOut :: erlang:timeout(),
+      Result  :: updated() | selected(),
+      Reason  :: process_not_owner_of_odbc_connection | common_reason().
+%%
+%%--------------------------------------------------------------------------
 sql_query(ConnectionReference, SQLQuery, infinity) when 
   is_pid(ConnectionReference), is_list(SQLQuery) -> 
     ODBCCmd = [?QUERY, SQLQuery],
@@ -194,17 +258,27 @@ sql_query(ConnectionReference, SQLQuery, TimeOut)
     call(ConnectionReference, {sql_query, ODBCCmd}, TimeOut).
 
 %%--------------------------------------------------------------------------
-%% select_count(ConnectionReference, SQLQuery, <TimeOut>) -> {ok, NrRows} |
-%%							    {error, Reason} 
+-spec select_count(ConnectionReference, SQLQuery) -> {ok, NrRows} |
+          {error, Reason} when
+      ConnectionReference :: connection_reference(),
+      SQLQuery :: string(),
+      NrRows :: n_rows(),
+      Reason :: process_not_owner_of_odbc_connection | common_reason().
 %%                                    
-%% Description: Executes a SQL SELECT query and associates the result set
-%%              with the connection. A cursor is positioned before
-%%        	the first row in the result set and the number of
-%%	        rows in the result set is returned.
 %%--------------------------------------------------------------------------
 select_count(ConnectionReference, SQLQuery) ->	
     select_count(ConnectionReference, SQLQuery, ?DEFAULT_TIMEOUT).
 
+%%--------------------------------------------------------------------------
+-spec select_count(ConnectionReference, SQLQuery, TimeOut) -> {ok, NrRows} |
+          {error, Reason} when
+      ConnectionReference :: connection_reference(),
+      SQLQuery :: string(),
+      TimeOut :: erlang:timeout(),
+      NrRows :: n_rows(),
+      Reason :: process_not_owner_of_odbc_connection | common_reason().
+%%
+%%--------------------------------------------------------------------------
 select_count(ConnectionReference, SQLQuery, infinity) when 
   is_pid(ConnectionReference), is_list(SQLQuery) ->
     ODBCCmd = [?SELECT_COUNT, SQLQuery],
@@ -216,15 +290,25 @@ select_count(ConnectionReference, SQLQuery, TimeOut) when
     call(ConnectionReference, {select_count, ODBCCmd}, TimeOut).
 
 %%--------------------------------------------------------------------------
-%% first(ConnectionReference, <TimeOut>) ->  {selected, ColNames, Rows} | 
-%%					     {error, Reason} 
-%%                                    
-%% Description: Selects the first row in the current result set. The cursor
-%%            : is positioned at this row. 
+-spec first(ConnectionReference) ->  Result | {error, Reason}  when
+      ConnectionReference :: connection_reference(),
+      Result :: selected(),
+      Reason :: result_set_does_not_exist | driver_does_not_support_function |
+                scrollable_cursors_disabled | process_not_owner_of_odbc_connection |
+                common_reason().
 %%--------------------------------------------------------------------------
 first(ConnectionReference) ->	
     first(ConnectionReference, ?DEFAULT_TIMEOUT).	
 
+%%--------------------------------------------------------------------------
+-spec first(ConnectionReference, TimeOut) -> Result | {error, Reason} when
+      ConnectionReference :: connection_reference(),
+      TimeOut :: erlang:timeout(),
+      Result :: selected(),
+      Reason :: result_set_does_not_exist | driver_does_not_support_function |
+                scrollable_cursors_disabled | process_not_owner_of_odbc_connection |
+                common_reason().
+%%--------------------------------------------------------------------------
 first(ConnectionReference, infinity) when is_pid(ConnectionReference) ->	
     ODBCCmd = [?SELECT, ?SELECT_FIRST],
     call(ConnectionReference, {select_cmd, absolute, ODBCCmd}, infinity);
@@ -235,15 +319,25 @@ first(ConnectionReference, TimeOut)
     call(ConnectionReference, {select_cmd, absolute, ODBCCmd}, TimeOut).
 
 %%--------------------------------------------------------------------------
-%% last(ConnectionReference, <TimeOut>) -> {selected, ColNames, Rows} | 
-%%					   {error, Reason} 
-%%                                    
-%% Description: Selects the last row in the current result set. The cursor
-%%            : is positioned at this row. 
+-spec last(ConnectionReference) -> Result | {error, Reason} when
+      ConnectionReference :: connection_reference(),
+      Result :: selected(),
+      Reason :: result_set_does_not_exist | driver_does_not_support_function |
+                scrollable_cursors_disabled | process_not_owner_of_odbc_connection |
+                common_reason().
 %%--------------------------------------------------------------------------
 last(ConnectionReference) ->	
     last(ConnectionReference, ?DEFAULT_TIMEOUT).	
 
+%%--------------------------------------------------------------------------
+-spec last(ConnectionReference, TimeOut) -> Result | {error, Reason} when
+      ConnectionReference :: connection_reference(),
+      TimeOut :: erlang:timeout(),
+      Result :: selected(),
+      Reason :: result_set_does_not_exist | driver_does_not_support_function |
+                scrollable_cursors_disabled | process_not_owner_of_odbc_connection |
+                common_reason().
+%%--------------------------------------------------------------------------
 last(ConnectionReference, infinity) when is_pid(ConnectionReference) ->	
     ODBCCmd = [?SELECT, ?SELECT_LAST],
     call(ConnectionReference, {select_cmd, absolute, ODBCCmd}, infinity);
@@ -252,17 +346,27 @@ last(ConnectionReference, TimeOut)
   when is_pid(ConnectionReference), is_integer(TimeOut), TimeOut > 0 ->	
     ODBCCmd = [?SELECT, ?SELECT_LAST],
     call(ConnectionReference, {select_cmd, absolute, ODBCCmd}, TimeOut).
+
 %%--------------------------------------------------------------------------
-%% next(ConnectionReference, <TimeOut>) -> {selected, ColNames, Rows} | 
-%%					   {error, Reason}  
-%%                                    
-%% Description: Selects the next row relative the current cursor position 
-%%            : in the current result set. The cursor is positioned at 
-%%            : this row. 
+-spec next(ConnectionReference) -> Result | {error, Reason} when
+      ConnectionReference :: connection_reference(),
+      Result :: selected(),
+      Reason :: result_set_does_not_exist | driver_does_not_support_function |
+                scrollable_cursors_disabled | process_not_owner_of_odbc_connection |
+                common_reason().
 %%--------------------------------------------------------------------------
 next(ConnectionReference) ->	
     next(ConnectionReference, ?DEFAULT_TIMEOUT).	
-    
+
+%%--------------------------------------------------------------------------
+-spec next(ConnectionReference, TimeOut) -> Result | {error, Reason} when
+     ConnectionReference :: connection_reference(),
+      TimeOut :: erlang:timeout(),
+      Result :: selected(),
+      Reason :: result_set_does_not_exist | driver_does_not_support_function |
+                scrollable_cursors_disabled | process_not_owner_of_odbc_connection |
+                common_reason().
+%%--------------------------------------------------------------------------
 next(ConnectionReference, infinity) when is_pid(ConnectionReference) ->	
     ODBCCmd = [?SELECT, ?SELECT_NEXT],
     call(ConnectionReference, {select_cmd, next, ODBCCmd}, infinity);
@@ -273,16 +377,25 @@ next(ConnectionReference, TimeOut)
     call(ConnectionReference, {select_cmd, next, ODBCCmd}, TimeOut).
 
 %%--------------------------------------------------------------------------
-%% prev(ConnectionReference, <TimeOut>) -> {selected, ColNames, Rows} | 
-%%					   {error, Reason}   
-%%                                    
-%% Description: Selects the previous row relative the current cursor 
-%%            : position in the current result set. The cursor is
-%%            : positioned at this row. 
+-spec prev(ConnectionReference) -> Result | {error, Reason} when
+      ConnectionReference :: connection_reference(),
+      Result :: selected(),
+      Reason :: result_set_does_not_exist | driver_does_not_support_function |
+                scrollable_cursors_disabled | process_not_owner_of_odbc_connection |
+                common_reason().
 %%--------------------------------------------------------------------------
 prev(ConnectionReference) ->	
     prev(ConnectionReference, ?DEFAULT_TIMEOUT).	
 
+%%--------------------------------------------------------------------------
+-spec prev(ConnectionReference, TimeOut) -> Result | {error, Reason} when
+      ConnectionReference :: connection_reference(),
+      TimeOut :: erlang:timeout(),
+      Result :: selected(),
+      Reason :: result_set_does_not_exist | driver_does_not_support_function |
+                scrollable_cursors_disabled | process_not_owner_of_odbc_connection |
+                common_reason().
+%%--------------------------------------------------------------------------
 prev(ConnectionReference, infinity) when is_pid(ConnectionReference) ->	
     ODBCCmd = [?SELECT, ?SELECT_PREV],
     call(ConnectionReference, {select_cmd, relative, ODBCCmd}, infinity);
@@ -293,21 +406,29 @@ prev(ConnectionReference, TimeOut)
     call(ConnectionReference, {select_cmd, relative, ODBCCmd}, TimeOut).
 
 %%--------------------------------------------------------------------------
-%% select(ConnectionReference, <Timeout>) -> {selected, ColNames, Rows} | 
-%%					     {error, Reason}   
-%%                                   
-%% Description: Selects <N> rows. If <Position> is next it is
-%%              semanticly eqvivivalent of calling next/[1,2] <N>
-%%              times. If <Position> is {relative, Pos} <Pos> will be
-%%              used as an offset from the current cursor position to
-%%              determine the first selected row. If <Position> is
-%%              {absolute, Pos}, <Pos> will be the number of the first
-%%              row selected. After this function has returned the
-%%              cursor is positioned at the last selected row.
+-spec select(ConnectionReference, Position, N) ->  Result | {error, Reason} when
+      ConnectionReference ::connection_reference(),
+      Position :: next | {relative, integer()} | {absolute, integer()},
+      N :: integer(),
+      Result  :: selected(),
+      Reason :: result_set_does_not_exist | driver_does_not_support_function
+              | scrollable_cursors_disabled | process_not_owner_of_odbc_connection |
+                common_reason().
 %%--------------------------------------------------------------------------
 select(ConnectionReference, Position, N) ->
     select(ConnectionReference, Position, N, ?DEFAULT_TIMEOUT).
 
+%%--------------------------------------------------------------------------
+-spec select(ConnectionReference, Position, N, TimeOut) ->  Result | {error, Reason} when
+      ConnectionReference ::connection_reference(),
+      Position :: next | {relative, integer()} | {absolute, integer()},
+      N :: integer(),
+      TimeOut :: erlang:timeout(),
+      Result  :: selected(),
+      Reason :: result_set_does_not_exist | driver_does_not_support_function
+              | scrollable_cursors_disabled | process_not_owner_of_odbc_connection |
+                common_reason().
+%%--------------------------------------------------------------------------
 select(ConnectionReference, next, N, infinity) 
   when is_pid(ConnectionReference), is_integer(N), N > 0 ->
     ODBCCmd = [?SELECT, ?SELECT_N_NEXT,
@@ -355,14 +476,31 @@ select(ConnectionReference, {absolute, Pos} , N, TimeOut)
     call(ConnectionReference, {select_cmd, absolute, ODBCCmd},
 	 TimeOut).
 %%--------------------------------------------------------------------------
-%% param_query(ConnectionReference, SQLQuery, Params, <TimeOut>) -> 
-%%                             ok | {error, Reason} 
-%%                                    
-%% Description: Executes a parameterized update/delete/insert-query. 
+-spec param_query(ConnectionReference, SQLQuery, Params) ->
+          Result | {error, Reason} when
+      ConnectionReference :: connection_reference(),
+      SQLQuery :: string(),
+      Params   :: [{odbc_data_type(), [value()]}] |[{odbc_data_type(), in | out| inout, [value()]}],
+      Result ::  selected() | updated(),
+      Reason :: driver_does_not_support_function |
+                process_not_owner_of_odbc_connection |
+                common_reason().
 %%--------------------------------------------------------------------------
 param_query(ConnectionReference, SQLQuery, Params) ->
     param_query(ConnectionReference, SQLQuery, Params, ?DEFAULT_TIMEOUT).
 
+%%--------------------------------------------------------------------------
+-spec param_query(ConnectionReference, SQLQuery, Params, TimeOut) ->
+          Result | {error, Reason} when
+      ConnectionReference :: connection_reference(),
+      SQLQuery :: string(),
+      Params   :: [{odbc_data_type(), [value()]}] |[{odbc_data_type(), in | out| inout, [value()]}],
+      TimeOut :: erlang:timeout(),
+      Result ::  selected() | updated(),
+      Reason :: driver_does_not_support_function |
+                process_not_owner_of_odbc_connection |
+                common_reason().
+%%--------------------------------------------------------------------------
 param_query(ConnectionReference, SQLQuery, Params, infinity) 
   when is_pid(ConnectionReference), is_list(SQLQuery), is_list(Params) ->
     Values = param_values(Params),
@@ -383,17 +521,27 @@ param_query(ConnectionReference, SQLQuery, Params, TimeOut)
     call(ConnectionReference, {param_query, ODBCCmd}, TimeOut).
 
 %%--------------------------------------------------------------------------
-%% describe_table(ConnectionReference, Table, <TimeOut>) -> {ok, Desc} 
-%%
-%% Desc - [{ColName, Datatype}]
-%% ColName - atom()
-%% Datatype - atom()                                    
-%% Description: Queries the database to find out the datatypes of the
-%%              table <Table>
+-spec describe_table(ConnectionReference, Table) ->
+          {ok, Description} | {error, Reason} when
+      ConnectionReference :: connection_reference(),
+      Table :: string(),
+      Description :: [{col_name(), odbc_data_type()}],
+      Reason ::  process_not_owner_of_odbc_connection | common_reason().
 %%--------------------------------------------------------------------------
+
 describe_table(ConnectionReference, Table) ->
     describe_table(ConnectionReference, Table, ?DEFAULT_TIMEOUT).
 
+%%--------------------------------------------------------------------------
+-spec describe_table(ConnectionReference, Table, TimeOut) ->
+          {ok, Description} | {error, Reason} when
+      ConnectionReference :: connection_reference(),
+      Table :: string(),
+      TimeOut :: erlang:timeout(),
+      Description :: [{col_name(), odbc_data_type()}],
+      Reason ::  process_not_owner_of_odbc_connection | common_reason().
+%%--------------------------------------------------------------------------
+
 describe_table(ConnectionReference, Table, infinity) when 
   is_pid(ConnectionReference), is_list(Table) -> 
     ODBCCmd = [?DESCRIBE, "SELECT * FROM " ++ Table],
